/*
 * aligner.cpp
 *
 *  Created on: 2010/09/28
 *      Author: shu
 */

#include "chain_filter.h"
#include "aligner_parameters.h"
#include "encoder.h"
#include "edit_blocks.h"
#include "queries.h"
#include "db.h"
#include "searcher.h"
#include "statistics.h"
#include "ungapped_extender.h"
#include "gapped_extender.h"
#include "logger.h"
#include "timer.h"
#include <fstream>
#include <iostream>
#include <vector>
#include <algorithm>
#include <stdint.h>
#include <utility>
#include <sstream>
#include <assert.h>
#include <omp.h>
#include "aligner.h"

using namespace std;

Aligner::Aligner() :
		kNextQueryIdIncrement_(1), kNextAlignmentIdIncrement_(1), finished_(0), next_query_id_(
				0), next_alignment_id_(0) {
}

Aligner::~Aligner() {
}

void Aligner::Run(AlignerParameters &parameters) {
	kNextQueryIdIncrement_ = parameters.GetkNextQueryIdIncrement();
	kNextAlignmentIdIncrement_ = parameters.GetkNextAlignmentIdIncrement();
	ofstream out(parameters.GetOutputFileName().c_str());
	timer::Timer t;
	Queries queries(parameters);
	Db db(parameters.GetDbFileName());
	vector<vector<Alignment> > best_alignments(0);
	Logger *logger = Logger::GetInstance();

	if (!out) {
		throw invalid_argument("can't open output file.");
	}

	if (queries.GetNumberSequences() == 0) {
		throw invalid_argument("there are no queries.");
	}

	for (; queries.GetNumberSequences() != 0; queries.Next()) {
		db.ResetChunk();
		db.GetIndex();
		db.GetConcatenatedSequence();
		logger->Log(">Presearching ...");
		t.restart();
		Presearch(parameters, queries, db, best_alignments);
		{
			stringstream m;
			m << " " << t << " sec.";
			logger->Log(m.str());
		}

		logger->Log(">Trace backing ...");
		t.restart();
		TraceBack(parameters, best_alignments, queries, db);
		{
			stringstream m;
			m << " " << t << " sec.";
			logger->Log(m.str());
		}

		logger->Log(">Writing output ...");
		t.restart();
		WriteOutput(parameters, out, best_alignments, queries, db);
		{
			stringstream m;
			m << " " << t << " sec.";
			logger->Log(m.str());
		}
	}
	out.close();
	logger->Log("Complete!");
}

void *Aligner::RunPresearchThread(void *arg) {
	PresearchThreadArguments *p =
			reinterpret_cast<PresearchThreadArguments*>(arg);
	p->aligner->PresearchThread(p->thread_id, *(p->parameters), *(p->queries),
			*(p->db), *(p->best_alignments));
	return NULL;
}

void *Aligner::RunTraceBackThread(void *arg) {
	TraceBackThreadArguments *p =
			reinterpret_cast<TraceBackThreadArguments*>(arg);
	p->aligner->TraceBackThread(p->thread_id, *(p->parameters),
			*(p->best_alignments), *(p->alignments_id_list), *(p->queries),
			*(p->db));
	return NULL;
}

uint32_t Aligner::GetNextQueryId(Queries &queries) {
	uint32_t ret = 0;
#pragma omp critical(next_query_id_lock_)
	{
		ret = next_query_id_;
		if (ret < queries.GetNumberSequences()) {
			next_query_id_ += kNextQueryIdIncrement_;
		}
	}
	return ret;
}

void Aligner::ResetNextQueryId() {
#pragma omp critical(next_query_id_lock_)
	{
		next_query_id_ = 0;
	}
}

uint32_t Aligner::GetNextAlignmentId(
		vector<pair<uint32_t, uint32_t> > &alignments) {
	uint32_t ret = 0;
#pragma omp critical(next_alignment_id_lock_)
	{
		ret = next_alignment_id_;
		if (ret < alignments.size()) {
			next_alignment_id_ += kNextAlignmentIdIncrement_;
		}
	}
	return ret;
}

void Aligner::ResetNextAlignmentId() {
#pragma omp critical(next_alignment_id_lock_)
	{
		next_alignment_id_ = 0;
	}
}

void Aligner::Finished(const AlignerParameters &parameters) {
#pragma omp critical(finished_lock_)
	{
		if (finished_ < parameters.GetNumberThreads()) {
			++finished_;
		} else {
			assert(
					!"the number of function to call is over than the number of threads.");
		}
	}
}

int Aligner::GetFinished() {
	return finished_;
}

void Aligner::ResetFinished() {
#pragma omp critical(finished_lock_)
	{
		finished_ = 0;
	}
}

void Aligner::Presearch(AlignerParameters &parameters, Queries &queries, Db &db,
		vector<vector<Alignment> > &best_alignments) {
	const int MAX_N_THREADS = 32; //t.ishida should be modified...
	next_query_id_ = 0;
	finished_ = 0;
	best_alignments.clear();
	best_alignments.resize(queries.GetNumberSequences());

	assert(parameters.GetNumberThreads() < MAX_N_THREADS);
	omp_set_num_threads(parameters.GetNumberThreads());

	PresearchThreadArguments arg[MAX_N_THREADS];

	int i;
#pragma omp parallel private(i,arg)
	{
		i = omp_get_thread_num();
		arg[i].thread_id = i;
		arg[i].aligner = this;
		arg[i].parameters = &parameters;
		arg[i].queries = &queries;
		arg[i].db = &db;
		arg[i].best_alignments = &best_alignments;
		Aligner::RunPresearchThread(&arg[i]);
	}

}

void Aligner::PresearchThread(int thread_id, AlignerParameters &parameters,
		Queries &queries, Db &db, vector<vector<Alignment> > &best_alignments) {
	uint32_t query_id_start = 0;
	vector<Hit> hits;
	vector<vector<PositionEntry> > start_lists;
	vector<vector<PositionEntry> > end_lists;

	if (thread_id == 0) {
		ResetNextQueryId();
		db.ResetChunk();
	}
	Searcher searcher(parameters);
	ChainFilter chain_filter(parameters);
	GappedExtender gapped_extender;
#pragma omp barrier

	while (db.GetChunkId() < db.GetNumberChunks()) {
#pragma omp barrier
		start_lists.resize(db.GetNumberSequences());
		end_lists.resize(db.GetNumberSequences());

		while ((query_id_start = GetNextQueryId(queries))
				< queries.GetNumberSequences()) {
			uint32_t query_id_end = queries.GetNumberSequences();
			query_id_end = min(query_id_start + kNextQueryIdIncrement_,
					query_id_end);
			for (uint32_t query_id = query_id_start; query_id < query_id_end;
					++query_id) {
				Query *query = queries.GetQuery(query_id);
				searcher.Set(*query, db);
				while (1) {
					searcher.Search(hits);
					if (hits.empty()) {
						break;
					}
					chain_filter.Filter(*query, db, hits);
					for (vector<Hit>::iterator hits_it = hits.begin();
							hits_it != hits.end(); ++hits_it) {
						int onside_score = 0;
						int score = 0;
						int query_position;
						int database_position;
						hits_it->score = 0;
						gapped_extender.ExtendOneSide(
								query->GetSequence()
										+ hits_it->hit.queries_offset - 1,
								hits_it->hit.queries_offset - 1,
								db.GetConcatenatedSequence()
										+ hits_it->hit.db_offset - 1,
								AlignerParameters::kSequenceEnd, true,
								parameters.GetScoreMatrix(),
								parameters.GetOpenGap(),
								parameters.GetExtendGap(),
								parameters.GetGappedIdealExtendNominalCutoff(),
								&score, &query_position, &database_position,
								NULL);
						onside_score += score;
						hits_it->start.queries_offset =
								hits_it->hit.queries_offset - 1
										+ query_position;
						hits_it->start.db_offset = hits_it->hit.db_offset - 1
								+ database_position;
						hits_it->score = onside_score;
						onside_score = 0;

						gapped_extender.ExtendOneSide(
								query->GetSequence()
										+ hits_it->hit.queries_offset,
								query->GetSequenceLength()
										- hits_it->hit.queries_offset,
								db.GetConcatenatedSequence()
										+ hits_it->hit.db_offset,
								AlignerParameters::kSequenceEnd, false,
								parameters.GetScoreMatrix(),
								parameters.GetOpenGap(),
								parameters.GetExtendGap(),
								parameters.GetGappedIdealExtendNominalCutoff(),
								&score, &query_position, &database_position,
								NULL);
						onside_score += score;
						hits_it->score += onside_score;
						hits_it->end.queries_offset =
								hits_it->hit.queries_offset + query_position;
						hits_it->end.db_offset = hits_it->hit.db_offset
								+ database_position;

					}
					Merge(parameters, hits, db, start_lists, end_lists,
							best_alignments[query_id]);
				}
			}
		}
#pragma omp barrier
		if (db.GetNumberChunks() == 1)
			break;
		if (thread_id == 0) {
			ResetNextQueryId();
			db.NextChunk();
		}
#pragma omp barrier
	}
}

void Aligner::FillAlignmentData(Query &query, Db &db,
		vector<EditBlocks::EditOpType> &edits, Alignment *alignment) {
	Code *query_seq = query.GetSequence();
	Code *db_seq = db.GetConcatenatedSequence();
	uint32_t queries_offset = alignment->start.queries_offset;
	uint32_t db_offset = alignment->start.db_offset;
	uint32_t m = 0;
	uint32_t o = 0;
	uint32_t e = 0;
	EditBlocks::EditOpType prev_op = EditBlocks::kSubstitution;
	for (vector<EditBlocks::EditOpType>::iterator it = edits.begin();
			it != edits.end(); ++it) {
		switch (*it) {
		case EditBlocks::kSubstitution:
			if (query_seq[queries_offset] != db_seq[db_offset]) {
				++m;
			}
			++queries_offset;
			++db_offset;
			break;
		case EditBlocks::kInsertion:
			if (prev_op != EditBlocks::kInsertion) {
				++o;
			} else {
				++e;
			}
			++queries_offset;
			break;
		case EditBlocks::kDeletion:
			if (prev_op != EditBlocks::kDeletion) {
				++o;
			} else {
				++e;
			}
			++db_offset;
			break;
		default:
			assert(!"invalid edit op");
			break;
		}
		prev_op = *it;
	}
#if 0
	/// debug //////////////////////////////////////////////////
	cout << alignment->score << " "<<alignment->start.queries_offset << " "<< alignment->start.db_offset << " "<< alignment->end.queries_offset << " " << alignment->end.db_offset<<endl;
	/////////////////////////////////////////////////////////////
#endif
	alignment->subject_name = db.GetName(alignment->subject_id);
	alignment->alignment_length = edits.size();
	alignment->mismatches = m;
	alignment->gap_openings = o;
	alignment->gap_extendings = e;
	alignment->start.queries_offset = query.GetRealStart(
			alignment->start.queries_offset);
	alignment->end.queries_offset = query.GetRealEnd(
			alignment->end.queries_offset);
	alignment->start.db_offset -= (db.GetOffset(alignment->subject_id) - 1);
	alignment->end.db_offset -= (db.GetOffset(alignment->subject_id) - 1);
}

void Aligner::Merge(const AlignerParameters &parameters, vector<Hit> &hits,
		Db &db, vector<vector<PositionEntry> > &start_lists,
		vector<vector<PositionEntry> > &end_lists,
		vector<Alignment> &best_alignments) {
	vector<Alignment> new_best_alignments(0);
	sort(hits.begin(), hits.end(), HitComp());

#if 0
	// debug //////////////////////////////////////////////////
	cout << "sorted hits" << endl;
	for (uint32_t x = 0; x < hits.size(); ++x) {
		cout << hits[x].score << " "<<hits[x].start.queries_offset << " " << hits[x].start.db_offset << " " << hits[x].end.queries_offset << " " << hits[x].end.db_offset<< " " << hits[x].hit.queries_offset << " " << hits[x].hit.db_offset << endl;
	}
	cout << endl;
	///////////////////////////////////////////////////////////
#endif 

	vector<Alignment>::iterator best_alignments_it = best_alignments.begin();
	vector<Alignment>::iterator best_alignments_end = best_alignments.end();
	vector<Hit>::iterator hits_it = hits.begin();
	vector<Hit>::iterator hits_end = hits.end();
	while (new_best_alignments.size() < parameters.GetMaxNumberAlignments()
			&& (hits_it != hits_end || best_alignments_it != best_alignments_end)) {
		Alignment added_alignment;
		if (hits_it == hits_end) {
			added_alignment = *best_alignments_it;
			++best_alignments_it;
		} else if (best_alignments_it == best_alignments_end
				|| AlignmentComp(best_alignments_it->score,
						best_alignments_it->start, best_alignments_it->end,
						hits_it->score, hits_it->start, hits_it->end) < 0) {
			added_alignment.db_id = db.GetChunkId();
			added_alignment.subject_id = db.GetId(hits_it->start.db_offset);
			added_alignment.subject_name = "";
			added_alignment.score = hits_it->score;
			added_alignment.start = hits_it->start;
			added_alignment.end = hits_it->end;
			added_alignment.hit = hits_it->hit;
			++hits_it;
		} else {
			added_alignment = *best_alignments_it;
			++best_alignments_it;
		}
		if (added_alignment.db_id == db.GetChunkId()) {
			if (start_lists[added_alignment.subject_id].size()
					< parameters.GetMaxNumberOneSubjectAlignments()) {
				vector<PositionEntry> *start_list =
						&start_lists[added_alignment.subject_id];
				PositionEntry searched_entry;
				searched_entry.index = 0;
				searched_entry.position = added_alignment.start.queries_offset;
				vector<PositionEntry>::iterator start_list_it = lower_bound(
						start_list->begin(), start_list->end(), searched_entry,
						QueryPositionEntryComp());
				vector<PositionEntry>::iterator start_insert_it = start_list_it;

				vector<PositionEntry> *end_list =
						&end_lists[added_alignment.subject_id];
				searched_entry.index = 0;
				searched_entry.position = added_alignment.end.queries_offset;
				vector<PositionEntry>::iterator end_list_it = lower_bound(
						end_list->begin(), end_list->end(), searched_entry,
						QueryPositionEntryComp());
				vector<PositionEntry>::iterator end_insert_it = end_list_it;

				// check overlap
				bool added = true;
				for (;
						start_list_it != start_list->end()
								&& added_alignment.end.queries_offset
										< start_list_it->position;
						++start_list_it) {
					Alignment *alignment_ptr =
							&new_best_alignments[start_list_it->index];
					if (Overlapped(alignment_ptr->start, alignment_ptr->end,
							added_alignment.start, added_alignment.end) > 0) {
						added = false;
						break;
					}
				}

				if (added == true) {
					for (;
							end_list_it != end_list->end()
									&& added_alignment.end.queries_offset
											== end_list_it->position;
							++end_list_it) {
						Alignment *alignment_ptr =
								&new_best_alignments[end_list_it->index];
						if (Overlapped(alignment_ptr->start, alignment_ptr->end,
								added_alignment.start, added_alignment.end)
								> 0) {
							added = false;
							break;
						}
					}
				}

				if (added == true) {
					PositionEntry e;
					e.position = added_alignment.start.queries_offset;
					e.index = new_best_alignments.size();
					start_lists[added_alignment.subject_id].insert(
							start_insert_it, e);
					e.position = added_alignment.end.queries_offset;
					end_lists[added_alignment.subject_id].insert(end_insert_it,
							e);
					new_best_alignments.push_back(added_alignment);
				}
			}
		} else {
			new_best_alignments.push_back(added_alignment);
		}
	}
	best_alignments.clear();
	best_alignments.insert(best_alignments.begin(), new_best_alignments.begin(),
			new_best_alignments.end());
	for (vector<Alignment>::iterator it = new_best_alignments.begin();
			it != new_best_alignments.end(); ++it) {
		start_lists[it->subject_id].clear();
		end_lists[it->subject_id].clear();
	}
#if 0
	// debug //////////////////////////////////
	cout << "best_alignments" << endl;
	for (uint32_t x = 0; x < best_alignments.size(); ++x) {
		cout << best_alignments[x].score << " "<<best_alignments[x].start.queries_offset << " "<< best_alignments[x].start.db_offset << " "<< best_alignments[x].end.queries_offset << " " << best_alignments[x].end.db_offset<<endl;
	}
	//////////////////////////////////////////
#endif

}

void Aligner::TraceBack(AlignerParameters &parameters,
		vector<vector<Alignment> > &best_alignments, Queries &queries, Db &db) {
	const int MAX_N_THREADS = 32; //t.ishida, should be modified ...
	uint32_t number_chunks = db.GetNumberChunks();
	vector<vector<pair<uint32_t, uint32_t> > > alignments_id_list(
			number_chunks);
	for (uint32_t i = 0; i < queries.GetNumberSequences(); ++i) {
		for (uint32_t j = 0; j < best_alignments[i].size(); ++j) {
			alignments_id_list[best_alignments[i][j].db_id].push_back(
					make_pair(i, j));
		}
	}
	next_alignment_id_ = 0;
	TraceBackThreadArguments arg[MAX_N_THREADS];

	int i;
#pragma omp parallel private(i,arg)
	{
		i = omp_get_thread_num();
		arg[i].thread_id = i;
		arg[i].aligner = this;
		arg[i].parameters = &parameters;
		arg[i].queries = &queries;
		arg[i].db = &db;
		arg[i].alignments_id_list = &alignments_id_list;
		arg[i].best_alignments = &best_alignments;
		Aligner::RunTraceBackThread(&arg[i]);
	}

	for (uint32_t i = 0; i < queries.GetNumberSequences(); ++i) {
		sort(best_alignments[i].begin(), best_alignments[i].end(),
				FinalAlignmentComp());
	}
}

void Aligner::TraceBackThread(int thread_id, AlignerParameters &parameters,
		vector<vector<Alignment> > &best_alignments,
		vector<vector<pair<uint32_t, uint32_t> > > &alignments_id_list,
		Queries &queries, Db &db) {
	uint32_t alignments_id_start = 0;
	if (thread_id == 0) {
		db.ResetChunk();
		ResetNextAlignmentId();
	}
	EditBlocks temp_edit_blocks;
	EditBlocks edit_blocks;
	GappedExtender gapped_extender;
#pragma omp barrier

	while (db.GetChunkId() < db.GetNumberChunks()) {
#pragma omp barrier
		while ((alignments_id_start = GetNextAlignmentId(
				alignments_id_list[db.GetChunkId()]))
				< alignments_id_list[db.GetChunkId()].size()) {
			uint32_t alignments_id_end =
					alignments_id_list[db.GetChunkId()].size();
			alignments_id_end = min(
					alignments_id_start + kNextAlignmentIdIncrement_,
					alignments_id_end);
			for (uint32_t alignments_id = alignments_id_start;
					alignments_id < alignments_id_end; ++alignments_id) {
				pair<uint32_t, uint32_t> id =
						alignments_id_list[db.GetChunkId()][alignments_id];
				Alignment *a = &best_alignments[id.first][id.second];
				Query *query = queries.GetQuery(id.first);
				// debug /////////////////////////////////////////
				//	cout << a->score << " "<< a->start.queries_offset << " " << a->start.db_offset  << " " << a->end.queries_offset << " " << a->end.db_offset<< " " << a->hit.queries_offset << " " << a->hit.db_offset << endl;
				////////////////////////////////////////////////
				int score = 0;
				int query_position;
				int database_position;
				a->score = 0;
				edit_blocks.Clear();
				gapped_extender.ExtendOneSide(
						query->GetSequence() + a->hit.queries_offset - 1,
						a->hit.queries_offset - 1,
						db.GetConcatenatedSequence() + a->hit.db_offset - 1,
						AlignerParameters::kSequenceEnd, true,
						parameters.GetScoreMatrix(), parameters.GetOpenGap(),
						parameters.GetExtendGap(),
						parameters.GetFinalGappedIdealExtendNominalCutoff(),
						&score, &query_position, &database_position,
						&temp_edit_blocks);
				a->score += score;
				a->start.queries_offset = a->hit.queries_offset - 1
						+ query_position;
				a->start.db_offset = a->hit.db_offset - 1 + database_position;
				edit_blocks.Add(temp_edit_blocks);
				gapped_extender.ExtendOneSide(
						query->GetSequence() + a->hit.queries_offset,
						query->GetSequenceLength() - a->hit.queries_offset,
						db.GetConcatenatedSequence() + a->hit.db_offset,
						AlignerParameters::kSequenceEnd, false,
						parameters.GetScoreMatrix(), parameters.GetOpenGap(),
						parameters.GetExtendGap(),
						parameters.GetFinalGappedIdealExtendNominalCutoff(),
						&score, &query_position, &database_position,
						&temp_edit_blocks);
				a->score += score;
				a->end.queries_offset = a->hit.queries_offset + query_position;
				a->end.db_offset = a->hit.db_offset + database_position;
				edit_blocks.Add(temp_edit_blocks);
				vector<EditBlocks::EditOpType> edits = edit_blocks.ToVector();
				FillAlignmentData(*query, db, edits, a);
#if 0
				for (vector<EditBlocks::EditOpType>::iterator e = edits.begin(); e != edits.end(); ++e) {
					switch (*e) {
						case EditBlocks::kSubstitution:
						cerr << ".";
						break;
						case EditBlocks::kInsertion:
						cerr << "+";
						break;
						case EditBlocks::kDeletion:
						cerr << "-";
						break;
						default:
						assert(!"invalid edit op");
					}
				}
				cerr << endl << endl;
#endif
			}
		}
#pragma omp barrier
		if (db.GetNumberChunks() == 1)
			break;
		if (thread_id == 0) {
			db.NextChunk();
			ResetNextAlignmentId();
		}
#pragma omp barrier
	}
}

void Aligner::WriteOutput(const AlignerParameters &parameters, ostream &out,
		vector<vector<Alignment> > &best_alignments, Queries &queries, Db &db) {
	for (uint32_t i = 0; i < queries.GetNumberSequences(); ++i) {
		Query *query = queries.GetQuery(i);
		string query_name = query->GetName();
		Statistics statistics(*(parameters.GetAlignedType()));
		uint64_t search_space = statistics.CalculateSearchSpace(
				query->GetRealSequenceLength(), db.GetDBLenght(),
				db.GetNumberSequencesInDB(),
				parameters.GetGappedKarlinParameters(), parameters.GetAlpha(),
				parameters.GetBeta());
		for (vector<Alignment>::iterator it = best_alignments[i].begin();
				it != best_alignments[i].end(); ++it) {
			out << query_name << "\t";
			out << it->subject_name << "\t";
			out
					<< (1.0
							- (static_cast<float>(it->mismatches
									+ it->gap_openings + it->gap_extendings)
									/ static_cast<float>(it->alignment_length)))
							* 100 << "\t";
			out << it->alignment_length << "\t";
			out << it->mismatches << "\t";
			out << it->gap_openings << "\t";
			out << it->start.queries_offset << "\t";
			out << it->end.queries_offset << "\t";
			out << it->start.db_offset << "\t";
			out << it->end.db_offset << "\t";
			out
					<< Statistics::Nominal2EValue(it->score, search_space,
							parameters.GetGappedKarlinParameters()) << "\t";
			out
					<< Statistics::Nominal2Normalized(it->score,
							parameters.GetGappedKarlinParameters());
			//out << "(" << it->score << ")";
			out << '\n';
		}
	}
}

int Aligner::AlignmentComp(int a_score, Coordinate a_start, Coordinate a_end,
		int b_score, Coordinate b_start, Coordinate b_end) {
	if (a_score > b_score) {
		return 1;
	} else if (a_score < b_score) {
		return -1;
	}
	uint32_t a_q_length = a_end.queries_offset - a_start.queries_offset;
	uint32_t b_q_length = b_end.queries_offset - b_start.queries_offset;

	if (a_q_length > b_q_length) {
		return 1;
	} else if (a_q_length < b_q_length) {
		return -1;
	}

	uint32_t a_s_length = a_end.db_offset - a_start.db_offset;
	uint32_t b_s_length = b_end.db_offset - b_start.db_offset;

	if (a_s_length > b_s_length) {
		return 1;
	} else if (a_s_length < b_s_length) {
		return -1;
	}

	return 0;
}

bool Aligner::Overlapped(Coordinate a_start, Coordinate a_end,
		Coordinate b_start, Coordinate b_end) {
	if ((a_start.queries_offset == b_start.queries_offset
			&& a_start.db_offset == b_start.db_offset) // overlapped start
			|| (a_end.queries_offset == b_end.queries_offset
					&& a_end.db_offset == b_end.db_offset) // overlapped end
			|| (a_start.queries_offset <= b_start.queries_offset // a contain b
			&& b_end.queries_offset <= a_end.queries_offset
					&& a_start.db_offset <= b_start.db_offset
					&& b_end.db_offset <= a_end.db_offset)) {
		return true;
	} else {
		return false;
	}
}

