/*
 * index.h
 *
 *  Created on: 2009/10/19
 *      Author: shu
 */

#ifndef INDEX_H_
#define INDEX_H_

#include <iostream>
#include <limits.h>
#include <stdint.h>
#include <fstream>
#include "bit_vector.h"
#include "succinct_bit_vector.h"

class FmIndex {
public:

  FmIndex();
  FmIndex(uint8_t *text, uint32_t length, uint32_t log_interval);
  FmIndex(std::ifstream &is);

  virtual ~FmIndex() {
    Delete();
  }

  void Rebuild(uint8_t *text, uint32_t length, uint32_t log_interval);

  uint32_t Occur(uint8_t character, uint32_t position) {
    if (character >= range_)
      return 0;
    uint32_t code = codes_[character];
    if (code == UINT_MAX)
      return 0;

    if (position < primary_) {
      ++position;
    }
    uint32_t node = 0;
    for (;; code >>= 1) {
      uint32_t bit = (code & 1);
      if (bit == 1) {
        position = succint_bit_vectors_[node]->Rank1(position);
      } else {
        position = position - succint_bit_vectors_[node]->Rank1(position);
      }
      node = children_[2 * node + bit];
      if (position == 0 || IsLeaf(node)) {
        break;
      }
    }
    return position;
  }

  void Occur(uint32_t position, uint32_t *occured_list, uint8_t start_character, uint8_t end_character);

  uint32_t Next(uint32_t position, uint8_t *character) {
    uint32_t code;
    uint32_t node;
    uint32_t depth;

    if (position < primary_) {
      ++position;
    }

    code = 0;
    node = 0;
    depth = 0;
    for (;; ++depth) {
      uint32_t bit = bit_vectors_[node]->GetBit(position - 1);
      if (bit == 1) {
        position = succint_bit_vectors_[node]->Rank1(position);
        code = code | (1 << depth);
      } else {
        position = position - succint_bit_vectors_[node]->Rank1(position);
      }
      node = children_[2 * node + bit];
      if (position == 0 || IsLeaf(node)) {
        break;
      }
    }
    *character = decodes_[code];
    return position;
  }

  uint32_t Locate(uint32_t position);

  uint32_t GetPrimary() {
    return primary_;
  }

  uint32_t GetRange() {
    return range_;
  }

  uint32_t GetNumberAlphabet() {
    return number_alphabet_;
  }

  uint32_t GetLogInterval() {
    return log_interval_;
  }

  uint32_t GetAlphabetOffset(uint8_t c) {
    return alphabet_offset_[c];
  }

  uint32_t GetLength() {
    return length_;
  }

  void Load(std::ifstream &is);
  void Save(std::ofstream &os);

private:
  bool IsLeaf(uint32_t i) const {
    if (i >= number_alphabet_ - 1)
      return true;
    else
      return false;
  }

  uint32_t InitializeEdge(uint32_t u, uint32_t l, uint32_t r, uint32_t parents[]) {
    parents[u] = UINT_MAX;
    children_[2 * u] = l;
    children_[2 * u + 1] = r;
    parents[l] = u;
    parents[r] = u;
    return u;
  }

  void BuildTree();

  void SetDecodes();

  void SetNodesList();

  void Insert(uint32_t code, uint32_t* bit_vector_count);

  uint32_t primary_;
  uint32_t range_;
  uint32_t number_alphabet_;
  uint32_t log_interval_;
  uint32_t length_;
  uint32_t *nodes_depth_list_;
  uint32_t *nodes_code_list_;
  uint32_t* codes_;
  uint8_t* decodes_;
  uint32_t *alphabet_offset_;
  uint32_t* children_;
  uint32_t* positions_;
  BitVector ** bit_vectors_;
  SuccinctBitVector** succint_bit_vectors_;

  void Delete() {
    if (codes_ != NULL) {
      delete[] codes_;
    }
    codes_ = NULL;

    if (decodes_ != NULL) {
      delete[] decodes_;
    }
    decodes_ = NULL;

    if (alphabet_offset_ != NULL) {
      delete[] alphabet_offset_;
    }
    alphabet_offset_ = NULL;

    if (children_ != NULL) {
      delete[] children_;
    }
    children_ = NULL;

    if (positions_ != NULL) {
      delete[] positions_;
    }
    positions_ = NULL;

    if (nodes_code_list_ != NULL) {
      delete[] nodes_code_list_;
    }
    nodes_code_list_ = NULL;

    if (nodes_depth_list_ != NULL) {
      delete[] nodes_depth_list_;
    }
    nodes_depth_list_ = NULL;

    if (succint_bit_vectors_ != NULL) {
      for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
        delete succint_bit_vectors_[i];
      }
      delete[] succint_bit_vectors_;
    }
    succint_bit_vectors_ = NULL;

    if (bit_vectors_ != NULL) {
      for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
        delete bit_vectors_[i];
      }
      delete[] bit_vectors_;
    }
    bit_vectors_ = NULL;

  }

};

#endif /* INDEX_H_ */
